import tensorflow as tf


class TFAlexNet(tf.keras.Model):
    def __init__(self):
        super().__init__()
        self.features = tf.keras.Sequential([
            tf.keras.layers.ZeroPadding2D(2),
            tf.keras.layers.Conv2D(64, kernel_size=11, strides=4),
            tf.keras.layers.ReLU(),
            tf.keras.layers.MaxPool2D(pool_size=3, strides=2),
            tf.keras.layers.Conv2D(192, kernel_size=5, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.MaxPool2D(pool_size=3, strides=2),
            tf.keras.layers.Conv2D(384, kernel_size=3, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.Conv2D(256, kernel_size=3, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.Conv2D(256, kernel_size=3, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.MaxPool2D(pool_size=3, strides=2)
        ])

    def call(self, inputs):
        return self.features(inputs)


if __name__ == '__main__':
    inputs = tf.random.normal((1, 224, 224, 3))
    output = TFAlexNet()(inputs)
    shape = output.shape
    print(shape)
    assert shape[1:3] == [6, 6]



import torch
import torch.nn as nn
from fusionlab.segmentation.base import SegmentationModel
from fusionlab.utils import autopad


class UNet(SegmentationModel):
    def __init__(self, cin, num_cls, base_dim=64):
        """
        Base Unet
        Args:
            cin (int): input channels
            num_cls (int): number of classes
            base_dim (int): 1st stage dim of conv output
        """
        super().__init__()
        stage = 5
        self.encoder = Encoder(cin, base_dim=base_dim)
        self.bridger = Bridger()
        self.decoder = Decoder(cin=base_dim*(2**(stage-1)),
                               base_dim=base_dim*(2**(stage-2)))  # 1024, 512
        self.head = Head(base_dim, num_cls)


class Encoder(nn.Module):
    def __init__(self, cin, base_dim):
        """
        UNet Encoder
        Args:
            cin (int): input channels
            base_dim (int): 1st stage dim of conv output
        """
        super().__init__()
        self.pool = nn.MaxPool2d(2, 2)
        self.stage1 = BasicBlock(cin, base_dim)
        self.stage2 = BasicBlock(base_dim, base_dim * 2)
        self.stage3 = BasicBlock(base_dim * 2, base_dim * 4)
        self.stage4 = BasicBlock(base_dim * 4, base_dim * 8)
        self.stage5 = BasicBlock(base_dim * 8, base_dim * 16)

    def forward(self, x):
        s1 = self.stage1(x)
        x = self.pool(s1)
        s2 = self.stage2(x)
        x = self.pool(s2)
        s3 = self.stage3(x)
        x = self.pool(s3)
        s4 = self.stage4(x)
        x = self.pool(s4)
        s5 = self.stage5(x)

        return [s1, s2, s3, s4, s5]


class Decoder(nn.Module):
    def __init__(self, cin, base_dim):
        """
        Base UNet decoder
        Args:
            cin (int): input channels
            base_dim (int): output dim of deepest stage output
        """
        super().__init__()
        self.d4 = DecoderBlock(cin, cin//2, base_dim)
        self.d3 = DecoderBlock(base_dim, cin//4, base_dim//2)
        self.d2 = DecoderBlock(base_dim//2, cin//8, base_dim//4)
        self.d1 = DecoderBlock(base_dim//4, cin//16, base_dim//8)

    def forward(self, x):
        f1, f2, f3, f4, f5 = x
        x = self.d4(f5, f4)
        x = self.d3(x, f3)
        x = self.d2(x, f2)
        x = self.d1(x, f1)
        return x


class Bridger(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        outputs = [nn.Identity()(i) for i in x]
        return outputs


class Head(nn.Sequential):
    def __init__(self, cin, cout):
        """
        Basic conv head
        :param int cin: input channel
        :param int cout: output channel
        """
        conv = nn.Conv2d(cin, cout, 1)
        super().__init__(conv)


class BasicBlock(nn.Sequential):
    def __init__(self, cin, cout):
        conv1 = nn.Sequential(
            nn.Conv2d(cin, cout, 3, 1, autopad(3)),
            nn.ReLU(),
        )
        conv2 = nn.Sequential(
            nn.Conv2d(cout, cout, 3, 1, autopad(3)),
            nn.ReLU(),
        )
        super().__init__(conv1, conv2)


class DecoderBlock(nn.Module):
    def __init__(self, c1, c2, cout):
        """
        Base Unet decoder block for merging the outputs from 2 stages
        Args:
            c1: input dim of the deeper stage
            c2: input dim of the shallower stage
            cout: output dim of the block
        """
        super().__init__()
        self.up = nn.Upsample(scale_factor=2)
        self.conv = BasicBlock(c1 + c2, cout)

    def forward(self, x1, x2):
        x1 = self.up(x1)
        x = torch.concat([x1, x2], dim=1)
        x = self.conv(x)
        return x


if __name__ == '__main__':
    H = W = 224
    dim = 64
    inputs = torch.normal(0, 1, (1, 3, H, W))

    encoder = Encoder(3, base_dim=dim)
    outputs = encoder(inputs)
    for i, o in enumerate(outputs):
        assert list(o.shape) == [1, dim*(2**i), H//(2**i), W//(2**i)]

    bridger = Bridger()
    outputs = bridger(outputs)
    for i, o in enumerate(outputs):
        assert list(o.shape) == [1, dim * (2 ** i), H // (2 ** i), W // (2 ** i)]

    features = [torch.normal(0, 1, (1, dim * (2 ** i), H // (2 ** i), W // (2 ** i))) for i in range(5)]
    decoder = Decoder(1024, 512)
    outputs = decoder(features)
    assert list(outputs.shape) == [1, 64, H, W]

    head = Head(64, 10)
    outputs = head(outputs)
    assert list(outputs.shape) == [1, 10, H, W]

    unet = UNet(3, 10)
    outputs = unet(inputs)
    assert list(outputs.shape) == [1, 10, H, W]



# kernel, padding
def autopad(kernal_size, padding=None):
    # Pad to 'same'
    if padding is None:
        padding = kernal_size//2 if isinstance(kernal_size, int) else [x//2 for x in kernal_size]
    return padding


import torch.nn as nn
from tensorflow.keras import Model


class SegmentationModel(nn.Module):
    """
    Base PyTorch class of the segmentation model with Encoder, Bridger, Decoder, Head
    """
    def forward(self, x):
        features = self.encoder(x)
        feature_fusion = self.bridger(features)
        decoder_output = self.decoder(feature_fusion)
        output = self.head(decoder_output)
        return output


class TFSegmentationModel(Model):
    """
    Base PyTorch class of the segmentation model with Encoder, Bridger, Decoder, Head
    """
    def call(self, x, training=None):
        """

        Args:
            x: input tensor
            training: flag for BatchNormalization and Dropout, whether the layer should behave in training mode or in
            inference mode

        Returns:

        """
        features = self.encoder(x, training)
        feature_fusion = self.bridger(features, training)
        decoder_output = self.decoder(feature_fusion, training)
        output = self.head(decoder_output, training)
        return output



import torch
import torch.nn as nn

from fusionlab.utils import autopad

# ref: https://arxiv.org/abs/1409.4842
# Going Deeper with Convolutions
class ConvBlock(nn.Module):
    def __init__(self, cin, cout, kernel_size=3, stride=1):
        super().__init__()
        self.conv = nn.Conv2d(cin, cout, kernel_size, stride, padding=autopad(kernel_size))
        self.act = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv(x)
        x = self.act(x)
        return x


class InceptionBlock(nn.Module):
    def __init__(self, cin, dim0, dim1, dim2, dim3):
        super().__init__()
        self.branch1 = ConvBlock(cin, dim0)
        self.branch3 = nn.Sequential(ConvBlock(cin, dim1[0], 1),
                                     ConvBlock(dim1[0], dim1[1], 3))
        self.branch5 = nn.Sequential(ConvBlock(cin, dim2[0], 1),
                                     ConvBlock(dim2[0], dim2[1], 5))
        self.pool = nn.Sequential(nn.MaxPool2d(3, 1, autopad(3)),
                                  ConvBlock(cin, dim3))

    def forward(self, x):
        x0 = self.branch1(x)
        x1 = self.branch3(x)
        x2 = self.branch5(x)
        x3 = self.pool(x)
        x = torch.cat((x0, x1, x2, x3), 1)
        return x


class InceptionNetV1(nn.Module):
    def __init__(self, cin=3):
        super().__init__()
        self.stem = nn.Sequential(
            ConvBlock(cin, 64, 7, stride=2),
            nn.MaxPool2d(3, 2, padding=autopad(3)),
            ConvBlock(64, 192, 3),
            nn.MaxPool2d(3, 2, padding=autopad(3)),
        )
        self.incept3a = InceptionBlock(192, 64, (96, 128), (16, 32), 32)
        self.incept3b = InceptionBlock(256, 128, (128, 192), (32, 96), 64)
        self.pool3 = nn.MaxPool2d(3, 2, padding=autopad(3))
        self.incept4a = InceptionBlock(480, 192, (96, 208), (16, 48), 64)
        self.incept4b = InceptionBlock(512, 160, (112, 224), (24, 64), 64)
        self.incept4c = InceptionBlock(512, 128, (128, 256), (24, 64), 64)
        self.incept4d = InceptionBlock(512, 112, (144, 288), (32, 64), 64)
        self.incept4e = InceptionBlock(528, 256, (160, 320), (32, 128), 128)
        self.pool4 = nn.MaxPool2d(3, 2, padding=autopad(3))
        self.incept5a = InceptionBlock(832, 256, (160, 320), (32, 128), 128)
        self.incept5b = InceptionBlock(832, 384, (192, 384), (48, 128), 128)

    def forward(self, x):
        x = self.stem(x)
        x = self.incept3a(x)
        x = self.incept3b(x)
        x = self.pool3(x)
        x = self.incept4a(x)
        x = self.incept4b(x)
        x = self.incept4c(x)
        x = self.incept4d(x)
        x = self.incept4e(x)
        x = self.pool4(x)
        x = self.incept5a(x)
        x = self.incept5b(x)
        return x

if __name__ == "__main__":
    inputs = torch.normal(0, 1, (1, 3, 224, 224))
    outputs = InceptionBlock(3, 64, (96, 128), (16, 32), 32)(inputs)
    print(outputs.shape)
    assert list(outputs.shape) == [1, 256, 224, 224]

    outputs = InceptionNetV1()(inputs)
    assert list(outputs.shape) == [1, 1024, 7, 7]



from .unet.unet import UNet
from .unet.tfunet import TFUNet
from .resunet.resunet import ResUNet
from .resunet.tfresunet import TFResUNet
from .unet2plus.unet2plus import UNet2plus
from .unet2plus.tfunet2plus import TFUNet2plus



import torch
import torch.nn as nn


class SEModule(nn.Module):
    def __init__(self, cin, ratio=16):
        super().__init__()
        cout = int(cin / ratio)
        self.gate = nn.Sequential(
            nn.Conv2d(cin, cout, kernel_size=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(cout, cin, kernel_size=1),
            nn.Sigmoid(),
        )

    def forward(self, inputs):
        x = inputs.mean((2, 3), keepdim=True)
        x = self.gate(x)
        return inputs * x


if __name__ == '__main__':
    inputs = torch.normal(0, 1, (1, 256, 224, 224))
    layer = SEModule(256)
    outputs = layer(inputs)
    assert list(outputs.shape) == [1, 256, 224, 224]



import torch
import torch.nn as nn


# Official pytorch ref: https://github.com/pytorch/vision/blob/main/torchvision/models/alexnet.py
class AlexNet(nn.Module):
    def __init__(self, c_in=3):
        super().__init__()
        self.features = nn.Sequential(
            nn.Conv2d(c_in, 64, kernel_size=11, stride=4, padding=2),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=3, stride=2),
            nn.Conv2d(64, 192, kernel_size=5, padding=2),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=3, stride=2),
            nn.Conv2d(192, 384, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(384, 256, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 256, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=3, stride=2),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.features(x)



if __name__ == '__main__':
    img_size = 224
    inputs = torch.normal(0, 1, (1, 3, img_size, img_size))
    output = AlexNet(3)(inputs)
    shape = list(output.shape)
    print(shape)
    assert shape[-2:] == [6, 6]



import tensorflow as tf
from tensorflow.keras import layers, Model, Sequential
from fusionlab.segmentation.base import TFSegmentationModel


class TFUNet(TFSegmentationModel):
    def __init__(self, num_cls, base_dim=64):
        """
        Base Unet
        Args:
            num_cls (int): number of classes
            base_dim (int): 1st stage dim of conv output
        """
        super().__init__()
        stage = 5
        self.encoder = Encoder(base_dim)
        self.bridger = Bridger()
        self.decoder = Decoder(base_dim*(2**(stage-2)))  # 512
        self.head = Head(num_cls)


class Encoder(Model):
    def __init__(self, base_dim):
        """
        UNet Encoder
        Args:
            base_dim (int): 1st stage dim of conv output
        """
        super().__init__()
        self.pool = layers.MaxPool2D()
        self.stage1 = BasicBlock(base_dim)
        self.stage2 = BasicBlock(base_dim * 2)
        self.stage3 = BasicBlock(base_dim * 4)
        self.stage4 = BasicBlock(base_dim * 8)
        self.stage5 = BasicBlock(base_dim * 16)

    def call(self, x, training=None):
        s1 = self.stage1(x, training)
        x = self.pool(s1)
        s2 = self.stage2(x, training)
        x = self.pool(s2)
        s3 = self.stage3(x, training)
        x = self.pool(s3)
        s4 = self.stage4(x, training)
        x = self.pool(s4)
        s5 = self.stage5(x, training)

        return [s1, s2, s3, s4, s5]


class Decoder(Model):
    def __init__(self, base_dim):
        """
        Base UNet decoder
        Args:
            base_dim (int): output dim of deepest stage output
        """
        super().__init__()
        self.d4 = DecoderBlock(base_dim)
        self.d3 = DecoderBlock(base_dim//2)
        self.d2 = DecoderBlock(base_dim//4)
        self.d1 = DecoderBlock(base_dim//8)

    def call(self, x, training=None):
        f1, f2, f3, f4, f5 = x
        x = self.d4(f5, f4, training)
        x = self.d3(x, f3, training)
        x = self.d2(x, f2, training)
        x = self.d1(x, f1, training)
        return x


class Bridger(Model):
    def __init__(self):
        super().__init__()

    def call(self, x, training=None):
        outputs = [tf.identity(i) for i in x]
        return outputs


class Head(Sequential):
    def __init__(self, cout):
        """
        Basic Identity
        :param int cout: output channel
        """
        conv = layers.Conv2D(cout, 1)
        super().__init__(conv)


class BasicBlock(Sequential):
    def __init__(self, cout):
        conv1 = Sequential([
            layers.Conv2D(cout, 3, 1, padding='same'),
            layers.ReLU(),
        ])
        conv2 = Sequential([
            layers.Conv2D(cout, 3, 1, padding='same'),
            layers.ReLU(),
        ])
        super().__init__([conv1, conv2])


class DecoderBlock(Model):
    def __init__(self, cout):
        """
        Base Unet decoder block for merging the outputs from 2 stages
        Args:
            cout: output dim of the block
        """
        super().__init__()
        self.up = layers.UpSampling2D()
        self.conv = BasicBlock(cout)

    def call(self, x1, x2, training=None):
        x1 = self.up(x1)
        x = tf.concat([x1, x2], axis=-1)
        x = self.conv(x, training)
        return x


if __name__ == '__main__':
    H = W = 224
    dim = 64
    inputs = tf.random.normal((1, H, W, 3))

    encoder = Encoder(dim)
    encoder.build((1, H, W, 3))
    outputs = encoder(inputs)
    for i, o in enumerate(outputs):
        assert list(o.shape) == [1, H // (2 ** i), W // (2 ** i), dim * (2 ** i)]

    bridger = Bridger()
    outputs = bridger(outputs)
    for i, o in enumerate(outputs):
        assert list(o.shape) == [1, H // (2 ** i), W // (2 ** i), dim * (2 ** i)]

    features = [tf.random.normal((1, H // (2 ** i), W // (2 ** i), dim * (2 ** i))) for i in range(5)]
    decoder = Decoder(512)
    outputs = decoder(features)
    assert list(outputs.shape) == [1, H, W, 64]

    head = Head(10)
    outputs = head(outputs)
    assert list(outputs.shape) == [1, H, W, 10]

    unet = TFUNet(10)
    outputs = unet(inputs)
    assert list(outputs.shape) == [1, H, W, 10]



import tensorflow as tf

__all__ = ["TFDiceLoss"]

BINARY_MODE = "binary"
MULTICLASS_MODE = "multiclass"

# class TFDiceCE(tf.keras.losses.Loss):
#     def __init__(self, cls_weight):
#         self.cls_weight = cls_weight
#         self.dice = DiceLoss()
#         self.ce = nn.CrossEntropyLoss(weight=cls_weight)

#     def forward(self, y_pred: Tensor, y_true: Tensor) -> Tensor:
#         loss_dice = self.dice(y_pred, y_true)
#         loss_ce = self.ce(y_pred, y_true)
#         return 0.5*loss_dice + 0.5*loss_ce


class TFDiceLoss(tf.keras.losses.Loss):
    def __init__(
        self,
        mode="binary",  # binary, multiclass
        log_loss=False,
        from_logits=True,
    ):
        """
        Implementation of Dice loss for image segmentation task.
        It supports "binary", "multiclass"
        https://github.com/BloodAxe/pytorch-toolbelt/blob/develop/pytorch_toolbelt/losses/dice.py
        Args:
            mode: Metric mode {'binary', 'multiclass'}
            log_loss: If True, loss computed as `-log(jaccard)`; otherwise `1 - jaccard`
            from_logits: If True assumes input is raw logits
        """
        super().__init__()
        self.mode = mode
        self.from_logits = from_logits
        self.log_loss = log_loss

    def call(self, y_pred, y_true):
        """
        :param y_pred: (N, H, W, C)
        :param y_true: (N, H, W)
        :return: scalar
        """

        y_true_shape = y_true.shape.as_list()
        y_pred_shape = y_pred.shape.as_list()
        assert y_true_shape[0] == y_pred_shape[0]
        # bs = y_true_shape[0]
        num_classes = y_pred_shape[-1]
        dims = [0, 1]  # (N, H*W)

        if self.from_logits:
            # get [0..1] class probabilities
            if self.mode == MULTICLASS_MODE:
                y_pred = tf.nn.softmax(y_pred, axis=-1)
            else:
                y_pred = tf.nn.sigmoid(y_pred)

        if self.mode == BINARY_MODE:
            y_true = tf.reshape(y_true, [-1, 1])
            y_pred = tf.reshape(y_pred, [-1, 1])
        elif self.mode == MULTICLASS_MODE:
            y_true = tf.reshape(y_true, [-1])
            y_true = tf.one_hot(y_true, num_classes)  # N, H*W -> N, H*W, C
            y_pred = tf.reshape(y_pred, [-1, num_classes])
        else:
            AssertionError("Not implemented")

        scores = soft_dice_score(y_pred, tf.cast(y_true, y_pred.dtype), dims=dims)
        if self.log_loss:
            loss = -tf.math.log(tf.clip_by_value(scores, clip_value_min=1e-7))
        else:
            loss = 1.0 - scores
        return tf.math.reduce_mean(loss)


def soft_dice_score(pred, target, dims=None):
    """
    Shape:
        - Input: :math:`(N, C, *)` where :math:`*` means any number of additional dimensions
        - Target: :math:`(N, C, *)`, same shape as the input
        - Output: scalar.
    """
    pred_shape = pred.shape.as_list()
    target_shape = target.shape.as_list()
    assert pred_shape == target_shape
    eps = 1e-7
    intersection = tf.reduce_sum(pred * target, axis=dims)
    cardinality = tf.reduce_sum(pred + target, axis=dims)
    dice_score = (2.0 * intersection) / tf.clip_by_value(cardinality, clip_value_min=eps, clip_value_max=cardinality.dtype.max)
    return dice_score


if __name__ == '__main__':
    pred = tf.convert_to_tensor([[
        [1., 2., 3.],
        [2., 6., 4.],
        [9., 6., 3.],
        [4., 8., 12.]
    ]])
    true = tf.convert_to_tensor([[2, 1, 0, 2]])
    dice = TFDiceLoss("multiclass", from_logits=True)
    loss = dice(pred, true)
    print(loss, "== 0.13497286?")

    print("Binary")
    pred = tf.convert_to_tensor([0.4, 0.2, 0.3, 0.5])
    pred = tf.reshape(pred, [1, 2, 2, 1])
    true = tf.convert_to_tensor([0, 1, 0, 1])
    true = tf.reshape(true, [1, 2, 2])

    print(pred.shape, true.shape)
    dice = TFDiceLoss("binary", from_logits=True)
    loss = dice(pred, true)
    print(loss, '== 0.4604469')


import tensorflow as tf
from tensorflow.keras import Model, layers, Sequential

# ref: https://arxiv.org/abs/1409.4842
# Going Deeper with Convolutions


class ConvBlock(Model):
    def __init__(self, cout, kernel_size=3, stride=1):
        super().__init__()
        self.conv = layers.Conv2D(cout, kernel_size, stride, padding="same")
        self.act = layers.ReLU()

    def call(self, inputs):
        x = self.conv(inputs)
        x = self.act(x)
        return x


class InceptionBlock(Model):
    def __init__(self, dim0, dim1, dim2, dim3):
        super().__init__()
        self.branch1 = ConvBlock(dim0, 1)
        self.branch3 = Sequential([
            ConvBlock(dim1[0], 1),
            ConvBlock(dim1[1], 3)
        ])
        self.branch5 = Sequential([
            ConvBlock(dim2[0], 1),
            ConvBlock(dim2[1], 5)]
        )
        self.pool = Sequential([
            layers.MaxPool2D(3, 1, padding='same'),
            ConvBlock(dim3, 1)
        ])

    def call(self, inputs):
        x = inputs
        x0 = self.branch1(x)
        x1 = self.branch3(x)
        x2 = self.branch5(x)
        x3 = self.pool(x)
        x = tf.concat([x0, x1, x2, x3], axis=-1)
        return x


class TFInceptionNetV1(Model):
    def __init__(self):
        super().__init__()
        self.stem = Sequential([
            ConvBlock(64, 7, stride=2),
            layers.MaxPool2D(3, 2, padding='same'),
            ConvBlock(192, 3),
            layers.MaxPool2D(3, 2, padding='same'),
        ])
        self.incept3a = InceptionBlock(64, (96, 128), (16, 32), 32)
        self.incept3b = InceptionBlock(128, (128, 192), (32, 96), 64)
        self.pool3 = layers.MaxPool2D(3, 2, padding='same')
        self.incept4a = InceptionBlock(192, (96, 208), (16, 48), 64)
        self.incept4b = InceptionBlock(160, (112, 224), (24, 64), 64)
        self.incept4c = InceptionBlock(128, (128, 256), (24, 64), 64)
        self.incept4d = InceptionBlock(112, (144, 288), (32, 64), 64)
        self.incept4e = InceptionBlock(256, (160, 320), (32, 128), 128)
        self.pool4 = layers.MaxPool2D(3, 2, padding='same')
        self.incept5a = InceptionBlock(256, (160, 320), (32, 128), 128)
        self.incept5b = InceptionBlock(384, (192, 384), (48, 128), 128)

    def call(self, inputs):
        x = self.stem(inputs)
        x = self.incept3a(x)
        x = self.incept3b(x)
        x = self.pool3(x)
        x = self.incept4a(x)
        x = self.incept4b(x)
        x = self.incept4c(x)
        x = self.incept4d(x)
        x = self.incept4e(x)
        x = self.pool4(x)
        x = self.incept5a(x)
        x = self.incept5b(x)
        return x


if __name__ == "__main__":
    inputs = tf.random.normal((1, 224, 224, 3))
    output = InceptionBlock(64, (96, 128), (16, 32), 32)(inputs)
    shape = output.shape
    print("InceptionBlock", shape)
    assert shape == (1, 224, 224, 256)

    output = TFInceptionNetV1()(inputs)
    shape = output.shape
    print("TFInceptionNetV1", shape)
    assert shape == (1, 7, 7, 1024)



import tensorflow as tf

# Official pytorch ref: https://github.com/pytorch/vision/blob/main/torchvision/models/vgg.py
class TFVGG16(tf.keras.Model):
    def __init__(self):
        super().__init__()
        ksize = 3
        self.features = tf.keras.Sequential([
            tf.keras.layers.Conv2D(64, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.Conv2D(64, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.MaxPool2D(),

            tf.keras.layers.Conv2D(128, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.Conv2D(128, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.MaxPool2D(),

            tf.keras.layers.Conv2D(256, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.Conv2D(256, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.Conv2D(256, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.MaxPool2D(),

            tf.keras.layers.Conv2D(512, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.Conv2D(512, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.Conv2D(512, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.MaxPool2D(),

            tf.keras.layers.Conv2D(512, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.Conv2D(512, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.Conv2D(512, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.MaxPool2D(),
        ])

    def call(self, inputs):
        return self.features(inputs)


class TFVGG19(tf.keras.Model):
    def __init__(self):
        super().__init__()
        ksize = 3
        self.features = tf.keras.Sequential([
            tf.keras.layers.Conv2D(64, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.Conv2D(64, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.MaxPool2D(),

            tf.keras.layers.Conv2D(128, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.Conv2D(128, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.MaxPool2D(),

            tf.keras.layers.Conv2D(256, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.Conv2D(256, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.Conv2D(256, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.Conv2D(256, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.MaxPool2D(),

            tf.keras.layers.Conv2D(512, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.Conv2D(512, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.Conv2D(512, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.Conv2D(512, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.MaxPool2D(),

            tf.keras.layers.Conv2D(512, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.Conv2D(512, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.Conv2D(512, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.Conv2D(512, ksize, padding='same'),
            tf.keras.layers.ReLU(),
            tf.keras.layers.MaxPool2D(),
        ])

    def call(self, inputs):
        return self.features(inputs)


if __name__ == '__main__':
    # VGG16
    inputs = tf.random.normal((1, 224, 224, 3))
    output = TFVGG16()(inputs)
    shape = output.shape
    assert shape[1:3] == [7, 7]

    # VGG19
    inputs = tf.random.normal((1, 224, 224, 3))
    output = TFVGG19()(inputs)
    shape = output.shape
    assert shape[1:3] == [7, 7]


import tensorflow as tf
from tensorflow.keras import layers, Model, Sequential
from fusionlab.segmentation.base import TFSegmentationModel


class TFUNet2plus(TFSegmentationModel):
    def __init__(self, num_cls, base_dim):
        super().__init__()
        self.encoder = Encoder(base_dim)
        self.bridger = Bridger()
        self.decoder = Decoder(base_dim)
        self.head = Head(num_cls)


class BasicBlock(Sequential):
    def __init__(self, cout):
        conv1 = Sequential([
            layers.Conv2D(cout, 3, 1, padding='same'),
            layers.BatchNormalization(),
            layers.Activation(tf.nn.relu),
        ])
        conv2 = Sequential([
            layers.Conv2D(cout, 3, 1, padding='same'),
            layers.BatchNormalization(),
            layers.Activation(tf.nn.relu),
        ])
        super().__init__([conv1, conv2])


class Encoder(Model):
    def __init__(self, base_dim):
        """
        UNet Encoder
        Args:
            base_dim (int): 1st stage dim of conv output
        """
        super().__init__()
        self.pool = layers.MaxPool2D()
        self.conv0_0 = BasicBlock(base_dim)
        self.conv1_0 = BasicBlock(base_dim * 2)
        self.conv2_0 = BasicBlock(base_dim * 4)
        self.conv3_0 = BasicBlock(base_dim * 8)
        self.conv4_0 = BasicBlock(base_dim * 16)

    def call(self, x, training):
        x0_0 = self.conv0_0(x, training)
        x1_0 = self.conv1_0(self.pool(x0_0), training)
        x2_0 = self.conv2_0(self.pool(x1_0), training)
        x3_0 = self.conv3_0(self.pool(x2_0), training)
        x4_0 = self.conv4_0(self.pool(x3_0), training)
        return [x0_0, x1_0, x2_0, x3_0, x4_0]


class Bridger(Model):
    def call(self, x, training=None):
        return [tf.identity(i) for i in x]


class Decoder(Model):
    def __init__(self, base_dim):
        super().__init__()
        dims = [base_dim*(2**i) for i in range(5)]  # [base_dim, base_dim*2, base_dim*4, base_dim*8, base_dim*16]
        self.conv0_1 = BasicBlock(dims[0])
        self.conv1_1 = BasicBlock(dims[1])
        self.conv2_1 = BasicBlock(dims[2])
        self.conv3_1 = BasicBlock(dims[3])

        self.conv0_2 = BasicBlock(dims[0])
        self.conv1_2 = BasicBlock(dims[1])
        self.conv2_2 = BasicBlock(dims[2])

        self.conv0_3 = BasicBlock(dims[0])
        self.conv1_3 = BasicBlock(dims[1])

        self.conv0_4 = BasicBlock(dims[0])
        self.up = layers.UpSampling2D()

    def call(self, x, training=None):
        x0_0, x1_0, x2_0, x3_0, x4_0 = x

        x0_1 = self.conv0_1(layers.concatenate([x0_0, self.up(x1_0)], -1))
        x1_1 = self.conv1_1(layers.concatenate([x1_0, self.up(x2_0)], -1))
        x0_2 = self.conv0_2(layers.concatenate([x0_0, x0_1, self.up(x1_1)], -1))

        x2_1 = self.conv2_1(layers.concatenate([x2_0, self.up(x3_0)], -1))
        x1_2 = self.conv1_2(layers.concatenate([x1_0, x1_1, self.up(x2_1)], -1))
        x0_3 = self.conv0_3(layers.concatenate([x0_0, x0_1, x0_2, self.up(x1_2)], -1))

        x3_1 = self.conv3_1(layers.concatenate([x3_0, self.up(x4_0)], -1))
        x2_2 = self.conv2_2(layers.concatenate([x2_0, x2_1, self.up(x3_1)], -1))
        x1_3 = self.conv1_3(layers.concatenate([x1_0, x1_1, x1_2, self.up(x2_2)], -1))
        x0_4 = self.conv0_4(layers.concatenate([x0_0, x0_1, x0_2, x0_3, self.up(x1_3)], -1))

        return x0_4


class Head(Sequential):
    def __init__(self, cout):
        """
        Basic Identity
        :param int cout: output channel
        """
        conv = layers.Conv2D(cout, 1)
        super().__init__(conv)


if __name__ == '__main__':
    H = W = 224
    dim = 32
    num_cls = 10
    inputs = tf.random.normal((1, H, W, 3))

    encoder = Encoder(base_dim=dim)
    outputs = encoder(inputs, training=True)
    for i, o in enumerate(outputs):
        assert list(o.shape) == [1, H // (2 ** i), W // (2 ** i), dim * (2 ** i)]

    bridger = Bridger()
    outputs = bridger(outputs, training=True)
    for i, o in enumerate(outputs):
        assert list(o.shape) == [1, H // (2 ** i), W // (2 ** i), dim * (2 ** i)]

    features = [tf.random.normal((1, H // (2 ** i), W // (2 ** i), dim * (2 ** i))) for i in range(5)]
    decoder = Decoder(dim)
    decoder.build([f.shape for f in features])
    outputs = decoder(features, training=True)
    assert list(outputs.shape) == [1, H, W, dim]

    head = Head(num_cls)
    outputs = head(outputs, training=True)
    assert list(outputs.shape) == [1, H, W, num_cls]

    unet = TFUNet2plus(num_cls, dim)
    outputs = unet(inputs, training=True)
    assert list(outputs.shape) == [1, H, W, num_cls]




import torch
import torch.nn as nn

from fusionlab.utils import autopad

# ResNet50
# Ref:
# https://github.com/keras-team/keras-applications/blob/master/keras_applications/resnet50.py

class ConvBlock(nn.Module):
    def __init__(self, cin, cout, kernel_size=3, stride=1, activation=True, padding=True):
        super().__init__()
        self.conv = nn.Conv2d(cin, cout, kernel_size, stride, autopad(kernel_size))
        self.bn = nn.BatchNorm2d(cout)
        self.act = nn.ReLU(inplace=True) if activation else nn.Identity()

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.act(x)
        return x


class Bottleneck(nn.Module):
    def __init__(self, cin, dims, kernel_size=3, stride=None):
        super().__init__()
        dim1, dim2 = dims
        self.conv1 = ConvBlock(cin, dim1, kernel_size=1)
        self.conv2 = ConvBlock(dim1, dim1, kernel_size=kernel_size,
                               stride=stride if stride else 1)
        self.conv3 = ConvBlock(dim1, dim2, kernel_size=1, activation=False)
        self.act = nn.ReLU(inplace=True)
        self.skip = nn.Identity() if not stride else ConvBlock(cin, dim2,
                                                               kernel_size=1,
                                                               stride=stride,
                                                               activation=False)

    def forward(self, x):
        identity = self.skip(x)

        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)

        x += identity
        x = self.act(x)
        return x


class Stem(nn.Sequential):
    def __init__(self, cin):
        super().__init__(
            ConvBlock(cin, 64, 7, stride=2),
            nn.MaxPool2d(3, 2, padding=autopad(3))
        )


class StageBlock(nn.Sequential):
    def __init__(self, cin, dims, stride, repeats):
        super().__init__()


class ResNet50V1(nn.Module):
    def __init__(self, cin):
        super().__init__()
        self.conv1 = Stem(cin)
        self.conv2 = nn.Sequential(
            Bottleneck(64, [64, 256], 3, stride=1),
            Bottleneck(256, [64, 256], 3),
            Bottleneck(256, [64, 256], 3),
        )
        self.conv3 = nn.Sequential(
            Bottleneck(256, [128, 512], 3, stride=2),
            Bottleneck(512, [128, 512], 3),
            Bottleneck(512, [128, 512], 3),
            Bottleneck(512, [128, 512], 3),
        )
        self.conv4 = nn.Sequential(
            Bottleneck(512, [256, 1024], 3, stride=2),
            Bottleneck(1024, [256, 1024], 3),
            Bottleneck(1024, [256, 1024], 3),
            Bottleneck(1024, [256, 1024], 3),
            Bottleneck(1024, [256, 1024], 3),
            Bottleneck(1024, [256, 1024], 3),
        )
        self.conv5 = nn.Sequential(
            Bottleneck(1024, [512, 2048], 3, stride=2),
            Bottleneck(2048, [512, 2048], 3),
            Bottleneck(2048, [512, 2048], 3),
        )
    def forward(self, x):
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.conv4(x)
        x = self.conv5(x)
        return x


if __name__ == '__main__':
    cin = 128
    inputs = torch.normal(0, 1, (1, cin, 224, 224))

    output = Bottleneck(cin, [64, 128])(inputs)
    shape = list(output.shape)
    print("Bottleneck", shape)
    assert shape == [1, 128, 224, 224]

    output = Bottleneck(cin, [128, 256], stride=1)(inputs)
    shape = list(output.shape)
    print("Bottleneck first conv for aligh dims", shape)
    assert shape == [1, 256, 224, 224]

    output = Bottleneck(cin, [64, 128], stride=2)(inputs)
    shape = list(output.shape)
    print("Bottleneck downsample", shape)
    assert shape == [1, 128, 112, 112]

    output = ResNet50V1(cin)(inputs)
    shape = list(output.shape)
    print("ResNet50V1", shape)
    assert shape == [1, 2048, 7, 7]







import tensorflow as tf
from tensorflow.keras import layers, Sequential


class TFSEModule(layers.Layer):
    def __init__(self, cin, ratio=16):
        super().__init__()
        cout = int(cin / ratio)
        self.gate = Sequential([
            layers.Conv2D(cout, kernel_size=1),
            layers.ReLU(),
            layers.Conv2D(cin, kernel_size=1),
            layers.Activation(tf.nn.sigmoid),
        ])

    def call(self, inputs):
        x = tf.reduce_mean(inputs, (1, 2), keepdims=True)
        x = self.gate(x)
        return inputs * x


if __name__ == '__main__':
    inputs = tf.random.normal((1, 224, 224, 256), 0, 1)
    layer = TFSEModule(256)
    outputs = layer(inputs)
    assert list(outputs.shape) == [1, 224, 224, 256]



import torch
from torch import nn
from fusionlab.segmentation.base import SegmentationModel
from fusionlab.utils import autopad

# ref: https://github.com/4uiiurz1/pytorch-nested-unet


class UNet2plus(SegmentationModel):
    def __init__(self, cin, num_cls, base_dim):
        super().__init__()
        self.encoder = Encoder(cin, base_dim)
        self.bridger = Bridger()
        self.decoder = Decoder(base_dim)
        self.head = Head(base_dim, num_cls)


class BasicBlock(nn.Sequential):
    def __init__(self, cin, cout):
        conv1 = nn.Sequential(
            nn.Conv2d(cin, cout, 3, 1, autopad(3)),
            nn.BatchNorm2d(cout),
            nn.ReLU(),
        )
        conv2 = nn.Sequential(
            nn.Conv2d(cout, cout, 3, 1, autopad(3)),
            nn.BatchNorm2d(cout),
            nn.ReLU(),
        )
        super().__init__(conv1, conv2)


class Encoder(nn.Module):
    def __init__(self, cin, base_dim):
        """
        UNet Encoder
        Args:
            cin (int): input channels
            base_dim (int): 1st stage dim of conv output
        """
        super().__init__()
        self.pool = nn.MaxPool2d(2, 2)
        self.conv0_0 = BasicBlock(cin, base_dim)
        self.conv1_0 = BasicBlock(base_dim, base_dim * 2)
        self.conv2_0 = BasicBlock(base_dim * 2, base_dim * 4)
        self.conv3_0 = BasicBlock(base_dim * 4, base_dim * 8)
        self.conv4_0 = BasicBlock(base_dim * 8, base_dim * 16)

    def forward(self, x):
        x0_0 = self.conv0_0(x)
        x1_0 = self.conv1_0(self.pool(x0_0))
        x2_0 = self.conv2_0(self.pool(x1_0))
        x3_0 = self.conv3_0(self.pool(x2_0))
        x4_0 = self.conv4_0(self.pool(x3_0))
        return [x0_0, x1_0, x2_0, x3_0, x4_0]


class Bridger(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return [nn.Identity()(i) for i in x]


class Decoder(nn.Module):
    def __init__(self, base_dim):
        super().__init__()
        dims = [base_dim*(2**i) for i in range(5)]  # [base_dim, base_dim*2, base_dim*4, base_dim*8, base_dim*16]
        self.conv0_1 = BasicBlock(dims[0] + dims[1], dims[0])
        self.conv1_1 = BasicBlock(dims[1] + dims[2], dims[1])
        self.conv2_1 = BasicBlock(dims[2] + dims[3], dims[2])
        self.conv3_1 = BasicBlock(dims[3] + dims[4], dims[3])

        self.conv0_2 = BasicBlock(dims[0] * 2 + dims[1], dims[0])
        self.conv1_2 = BasicBlock(dims[1] * 2 + dims[2], dims[1])
        self.conv2_2 = BasicBlock(dims[2] * 2 + dims[3], dims[2])

        self.conv0_3 = BasicBlock(dims[0] * 3 + dims[1], dims[0])
        self.conv1_3 = BasicBlock(dims[1] * 3 + dims[2], dims[1])

        self.conv0_4 = BasicBlock(dims[0] * 4 + dims[1], dims[0])
        self.up = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)

    def forward(self, x):
        x0_0, x1_0, x2_0, x3_0, x4_0 = x

        x0_1 = self.conv0_1(torch.cat([x0_0, self.up(x1_0)], 1))
        x1_1 = self.conv1_1(torch.cat([x1_0, self.up(x2_0)], 1))
        x0_2 = self.conv0_2(torch.cat([x0_0, x0_1, self.up(x1_1)], 1))

        x2_1 = self.conv2_1(torch.cat([x2_0, self.up(x3_0)], 1))
        x1_2 = self.conv1_2(torch.cat([x1_0, x1_1, self.up(x2_1)], 1))
        x0_3 = self.conv0_3(torch.cat([x0_0, x0_1, x0_2, self.up(x1_2)], 1))

        x3_1 = self.conv3_1(torch.cat([x3_0, self.up(x4_0)], 1))
        x2_2 = self.conv2_2(torch.cat([x2_0, x2_1, self.up(x3_1)], 1))
        x1_3 = self.conv1_3(torch.cat([x1_0, x1_1, x1_2, self.up(x2_2)], 1))
        x0_4 = self.conv0_4(torch.cat([x0_0, x0_1, x0_2, x0_3, self.up(x1_3)], 1))

        return x0_4


class Head(nn.Sequential):
    def __init__(self, cin, cout):
        """
        Basic Identity
        :param int cin: input channel
        :param int cout: output channel
        """
        conv = nn.Conv2d(cin, cout, 1)
        super().__init__(conv)


if __name__ == '__main__':
    H = W = 224
    dim = 32
    inputs = torch.normal(0, 1, (1, 3, H, W))

    encoder = Encoder(3, base_dim=dim)
    outputs = encoder(inputs)
    for i, o in enumerate(outputs):
        assert list(o.shape) == [1, dim * (2 ** i), H // (2 ** i), W // (2 ** i)]

    bridger = Bridger()
    outputs = bridger(outputs)
    for i, o in enumerate(outputs):
        assert list(o.shape) == [1, dim * (2 ** i), H // (2 ** i), W // (2 ** i)]

    features = [torch.normal(0, 1, (1, dim * (2 ** i), H // (2 ** i), W // (2 ** i))) for i in range(5)]
    decoder = Decoder(dim)
    outputs = decoder(features)
    assert list(outputs.shape) == [1, dim, H, W]

    head = Head(dim, 10)
    outputs = head(outputs)
    assert list(outputs.shape) == [1, 10, H, W]

    unet = UNet2plus(3, 10, dim)
    outputs = unet(inputs)
    assert list(outputs.shape) == [1, 10, H, W]



import torch
import torch.nn as nn


# Official pytorch ref: https://github.com/pytorch/vision/blob/main/torchvision/models/vgg.py
class VGG16(nn.Module):
    def __init__(self, c_in=3):
        super().__init__()
        ksize = 3
        self.features = nn.Sequential(
            nn.Conv2d(c_in, 64, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(64, 64, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),

            nn.Conv2d(64, 128, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(128, 128, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),

            nn.Conv2d(128, 256, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 256, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 256, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),

            nn.Conv2d(256, 512, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(512, 512, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(512, 512, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),

            nn.Conv2d(512, 512, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(512, 512, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(512, 512, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),
        )

    def forward(self, x):
        return self.features(x)


class VGG19(nn.Module):
    def __init__(self, c_in=3):
        super().__init__()
        ksize = 3
        self.features = nn.Sequential(
            nn.Conv2d(c_in, 64, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(64, 64, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),

            nn.Conv2d(64, 128, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(128, 128, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),

            nn.Conv2d(128, 256, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 256, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 256, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 256, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),

            nn.Conv2d(256, 512, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(512, 512, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(512, 512, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(512, 512, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),

            nn.Conv2d(512, 512, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(512, 512, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(512, 512, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(512, 512, ksize, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),
        )

    def forward(self, x):
        return self.features(x)




if __name__ == '__main__':
    # VGG16
    inputs = torch.normal(0, 1, (1, 3, 224, 224))
    output = VGG16(3)(inputs)
    shape = list(output.shape)
    assert shape[2:] == [7, 7]

    # VGG19
    inputs = torch.normal(0, 1, (1, 3, 224, 224))
    output = VGG19(3)(inputs)
    shape = list(output.shape)
    assert shape[2:] == [7, 7]


import torch
from torch import nn
import torch.nn.functional as F

__all__ = ["DiceLoss", "DiceCE"]

BINARY_MODE = "binary"
MULTICLASS_MODE = "multiclass"


class DiceCE(nn.Module):
    def __init__(self, w_dice=0.5, w_ce=0.5, cls_weight=None):
        """

        Args:
            w_dice: weight of Dice Loss
            w_ce: weight of CrossEntropy loss
            cls_weight:
        """
        super().__init__()
        self.w_dice = w_dice
        self.w_ce = w_ce
        self.cls_weight = cls_weight
        self.dice = DiceLoss()
        self.ce = nn.CrossEntropyLoss(weight=cls_weight)

    def forward(self, y_pred, y_true):
        loss_dice = self.dice(y_pred, y_true)
        loss_ce = self.ce(y_pred, y_true)
        return self.w_dice * loss_dice + self.w_ce * loss_ce


class DiceLoss(nn.Module):
    def __init__(
        self,
        mode="multiclass",  # binary
        log_loss=False,
        from_logits=True,
    ):
        """
        Implementation of Dice loss for image segmentation task.
        It supports "binary", "multiclass"
        https://github.com/BloodAxe/pytorch-toolbelt/blob/develop/pytorch_toolbelt/losses/dice.py
        Args:
            mode: Metric mode {'binary', 'multiclass'}
            log_loss: If True, loss computed as `-log(jaccard)`; otherwise `1 - jaccard`
            from_logits: If True assumes input is raw logits
        """
        super().__init__()
        self.mode = mode
        self.from_logits = from_logits
        self.log_loss = log_loss

    def forward(self, y_pred, y_true):
        """
        :param y_pred: (N, C, H, W)
        :param y_true: (N, H, W)
        :return: scalar
        """
        assert y_true.size(0) == y_pred.size(0)
        bs = y_true.size(0)
        num_classes = y_pred.size(1)
        dims = (0, 2) # # (N, -1, H*W)

        if self.from_logits:
            # get [0..1] class probabilities
            if self.mode == MULTICLASS_MODE:
                y_pred = F.softmax(y_pred, dim=1)
            else:
                y_pred = torch.sigmoid(y_pred)

        if self.mode == BINARY_MODE:
            y_true = y_true.view(bs, 1, -1)
            y_pred = y_pred.view(bs, 1, -1)
        elif self.mode == MULTICLASS_MODE:
            y_true = y_true.view(bs, -1)
            y_pred = y_pred.view(bs, num_classes, -1)

            y_true = F.one_hot(y_true, num_classes)  # N, H*W -> N,H*W, C
            y_true = y_true.permute(0, 2, 1)  # H, C, H*W
        else:
            AssertionError("Not implemented")

        scores = soft_dice_score(y_pred, y_true.type_as(y_pred), dims=dims)
        if self.log_loss:
            loss = -torch.log(scores.clamp_min(self.eps))
        else:
            loss = 1.0 - scores
        return loss.mean()


def soft_dice_score(pred, target, dims=None):
    """
    Shape:
        - Input: :math:`(N, C, *)` where :math:`*` means any number of additional dimensions
        - Target: :math:`(N, C, *)`, same shape as the input
        - Output: scalar.
    """
    assert pred.size() == target.size()
    eps = 1e-7
    intersection = torch.sum(pred * target, dim=dims)
    cardinality = torch.sum(pred + target, dim=dims)
    dice_score = (2.0 * intersection) / cardinality.clamp(min=eps)
    return dice_score


if __name__ == "__main__":
    pred = torch.tensor([[
        [1., 2., 3.],
        [2., 6., 4.],
        [9., 6., 3.],
        [4., 8., 12.]
    ]]).permute(0, 2, 1)
    true = torch.tensor([[2, 1, 0, 2]])
    print(pred.shape, true.shape)
    dice = DiceLoss("multiclass", from_logits=True)
    loss = dice(pred, true)
    print(loss, '== 0.1350?')

    print("Binary")
    pred = torch.tensor([0.4, 0.2, 0.3, 0.5]).reshape(1, 1, 2, 2)
    true = torch.tensor([0, 1, 0, 1]).reshape(1, 2, 2)
    print(pred.shape, true.shape)
    dice = DiceLoss("binary", from_logits=True)
    loss = dice(pred, true)
    print(loss, "== 0.4604?")




import tensorflow as tf
from tensorflow.keras import Model, Sequential, layers

# ResNet50
# Ref:
# https://github.com/keras-team/keras-applications/blob/master/keras_applications/resnet50.py
# https://github.com/raghakot/keras-resnet/blob/master/README.md


class Identity(layers.Layer):
    def __init__(self):
        super(Identity, self).__init__()

    def call(self, inputs, training=None):
        return inputs


class ConvBlock(Model):
    def __init__(self, cout, kernel_size=3, stride=1, activation=True, padding=True):
        super().__init__()
        self.conv = layers.Conv2D(cout, kernel_size, stride,
                                  padding='same' if padding else 'valid')
        self.bn = layers.BatchNormalization()
        self.act = layers.ReLU() if activation else Identity()

    def call(self, inputs, training=None):
        x = self.conv(inputs)
        x = self.bn(x, training)
        x = self.act(x)
        return x


class Bottleneck(Model):
    def __init__(self, dims, kernel_size=3, stride=None):
        super().__init__()
        dim1, dim2, dim3 = dims
        self.conv1 = ConvBlock(dim1, kernel_size=1)
        self.conv2 = ConvBlock(dim2, kernel_size=kernel_size,
                               stride=stride if stride else 1)
        self.conv3 = ConvBlock(dim3, kernel_size=1, activation=False)
        self.act = layers.ReLU()
        self.skip = Identity() if not stride else ConvBlock(dim3,
                                                            kernel_size=1,
                                                            stride=stride,
                                                            activation=False)

    def call(self, inputs, training=None):
        identity = self.skip(inputs, training)

        x = self.conv1(inputs, training)
        x = self.conv2(x, training)
        x = self.conv3(x, training)

        x += identity
        x = self.act(x)
        return x


class TFResNet50V1(Model):
    def __init__(self):
        super(TFResNet50V1, self).__init__()
        self.conv1 = Sequential([
            ConvBlock(64, 7, stride=2),
            layers.MaxPool2D(3, strides=2, padding='same'),
        ])
        self.conv2 = Sequential([
            Bottleneck([64, 64, 256], 3, stride=1),
            Bottleneck([64, 64, 256], 3),
            Bottleneck([64, 64, 256], 3),
        ])
        self.conv3 = Sequential([
            Bottleneck([128, 128, 512], 3, stride=2),
            Bottleneck([128, 128, 512], 3),
            Bottleneck([128, 128, 512], 3),
            Bottleneck([128, 128, 512], 3),
        ])
        self.conv4 = Sequential([
            Bottleneck([256, 256, 1024], 3, stride=2),
            Bottleneck([256, 256, 1024], 3),
            Bottleneck([256, 256, 1024], 3),
            Bottleneck([256, 256, 1024], 3),
            Bottleneck([256, 256, 1024], 3),
            Bottleneck([256, 256, 1024], 3),
        ])
        self.conv5 = Sequential([
            Bottleneck([512, 512, 2048], 3, stride=2),
            Bottleneck([512, 512, 2048], 3),
            Bottleneck([512, 512, 2048], 3),
        ])

    def call(self, inputs, training=None):
        x = self.conv1(inputs, training)
        x = self.conv2(x, training)
        x = self.conv3(x, training)
        x = self.conv4(x, training)
        x = self.conv5(x, training)
        return x


if __name__ == '__main__':
    inputs = tf.random.normal((1, 224, 224, 128))
    output = Bottleneck([64, 64, 128])(inputs)
    shape = output.shape
    print("Bottleneck", shape)
    assert shape == (1, 224, 224, 128)

    output = Bottleneck([128, 128, 256], stride=1)(inputs)
    shape = output.shape
    print("Bottleneck first conv for aligh dims", shape)
    assert shape == (1, 224, 224, 256)

    output = Bottleneck([64, 64, 128], stride=2)(inputs)
    shape = output.shape
    print("Bottleneck downsample", shape)
    assert shape == (1, 112, 112, 128)

    output = Identity()(inputs)
    shape = output.shape
    print("Identity", shape)
    assert shape == (1, 224, 224, 128)


    output = TFResNet50V1()(inputs)
    shape = output.shape
    print("TFResNet50V1", shape)
    assert  shape == (1, 7, 7, 2048)







import tensorflow as tf
from tensorflow.keras import layers, Model, Sequential
from fusionlab.segmentation.base import TFSegmentationModel


class TFResUNet(TFSegmentationModel):
    def __init__(self, num_cls, base_dim):
        super().__init__()
        self.encoder = Encoder(base_dim)
        self.bridger = Bridger()
        self.decoder = Decoder(base_dim)
        self.head = Head(num_cls)


class Encoder(Model):
    def __init__(self, base_dim):
        super().__init__()
        dims = [base_dim * (2 ** i) for i in range(4)]
        self.stem = Stem(dims[0])
        self.stage1 = ResConv(dims[1], stride=2)
        self.stage2 = ResConv(dims[2], stride=2)
        self.stage3 = ResConv(dims[3], stride=2)

    def call(self, x, training):
        s0 = self.stem(x, training)
        s1 = self.stage1(s0, training)
        s2 = self.stage2(s1, training)
        s3 = self.stage3(s2, training)
        return [s0, s1, s2, s3]


class Decoder(Model):
    def __init__(self, base_dim):
        """
        Base UNet decoder
        Args:
            base_dim (int): output dim of deepest stage output or input channels
        """
        super().__init__()
        dims = [base_dim*(2**i) for i in range(4)]
        self.d3 = DecoderBlock(dims[2])
        self.d2 = DecoderBlock(dims[1])
        self.d1 = DecoderBlock(dims[0])

    def call(self, x, training):
        s0, s1, s2, s3 = x

        x = self.d3(s3, s2, training)
        x = self.d2(x, s1, training)
        x = self.d1(x, s0, training)
        return x


class DecoderBlock(Model):
    def __init__(self, cout):
        super().__init__()
        self.upsample = layers.Conv2DTranspose(cout, 2, strides=2)
        self.conv = ResConv(cout, stride=1)

    def call(self, x1, x2, training):
        x1 = self.upsample(x1)
        x = tf.concat([x1, x2], axis=-1)
        return self.conv(x)


class Bridger(Model):
    def __init__(self):
        super().__init__()

    def call(self, x, training):
        outputs = [tf.identity(i) for i in x]
        return outputs


class Stem(Model):
    def __init__(self, cout):
        super().__init__()
        self.conv = Sequential([
            layers.Conv2D(cout, 3, padding='same'),
            layers.BatchNormalization(),
            layers.ReLU(),
            layers.Conv2D(cout, 3, padding='same'),
        ])
        self.skip = Sequential(
            layers.Conv2D(cout, 3, padding='same'),
        )

    def call(self, x, training):
        return self.conv(x) + self.skip(x)


class ResConv(Model):
    def __init__(self, cout, stride=1):
        super().__init__()

        self.conv = Sequential([
            layers.BatchNormalization(),
            layers.ReLU(),
            layers.Conv2D(cout, 3, stride, padding='same'),
            layers.BatchNormalization(),
            layers.ReLU(),
            layers.Conv2D(cout, 3, padding='same'),
        ])
        self.skip = Sequential([
            layers.Conv2D(cout, 3, strides=stride, padding='same'),
            layers.BatchNormalization(),
        ])

    def call(self, x, training=None):
        return self.conv(x, training) + self.skip(x, training)


class Head(Sequential):
    def __init__(self, cout):
        """
        Basic conv head
        :param int cout: number of classes
        """
        conv = layers.Conv2D(cout, 1)
        super().__init__(conv)


if __name__ == '__main__':
    H = W = 224
    cout = 64
    inputs = tf.random.normal((1, H, W, 3))

    model = TFResUNet(100, cout)
    output = model(inputs, training=True)
    print(output.shape)

    dblock = DecoderBlock(128)
    inputs2 = tf.random.normal((1, H, W, 128))
    inputs1 = tf.random.normal((1, H//2, W//2, 64))
    outputs = dblock(inputs1, inputs2, training=True)
    print(outputs.shape)

    encoder = Encoder(cout)
    outputs = encoder(inputs, training=True)
    for o in outputs:
        print(o.shape)

    decoder = Decoder(cout)
    outputs = decoder(outputs, training=True)
    print("Encoder + Decoder ", outputs.shape)

    stem = Stem(cout)
    outputs = stem(inputs, training=True)
    print(outputs.shape)
    assert list(outputs.shape) == [1, H, W, cout]

    resconv = ResConv(cout, 1)
    outputs = resconv(inputs, training=True)
    print(outputs.shape)
    assert list(outputs.shape) == [1, H, W, cout]

    resconv = ResConv(cout, stride=2)
    outputs = resconv(inputs, training=True)
    print(outputs.shape)
    assert list(outputs.shape) == [1, H//2, W//2, cout]


import torch
import torch.nn as nn
from fusionlab.segmentation.base import SegmentationModel
from fusionlab.utils import autopad


class ResUNet(SegmentationModel):
    def __init__(self, cin, num_cls, base_dim):
        super().__init__()
        self.encoder = Encoder(cin, base_dim)
        self.bridger = Bridger()
        self.decoder = Decoder(base_dim)
        self.head = Head(base_dim, num_cls)


class Encoder(nn.Module):
    def __init__(self, cin, base_dim):
        super().__init__()
        dims = [base_dim * (2 ** i) for i in range(4)]
        self.stem = Stem(cin, dims[0])
        self.stage1 = ResConv(dims[0], dims[1], stride=2)
        self.stage2 = ResConv(dims[1], dims[2], stride=2)
        self.stage3 = ResConv(dims[2], dims[3], stride=2)

    def forward(self, x):
        s0 = self.stem(x)
        s1 = self.stage1(s0)
        s2 = self.stage2(s1)
        s3 = self.stage3(s2)
        return [s0, s1, s2, s3]


class Decoder(nn.Module):
    def __init__(self, base_dim):
        """
        Base UNet decoder
        Args:
            base_dim (int): output dim of deepest stage output or input channels
        """
        super().__init__()
        dims = [base_dim*(2**i) for i in range(4)]
        self.d3 = DecoderBlock(dims[3], dims[2])
        self.d2 = DecoderBlock(dims[2], dims[1])
        self.d1 = DecoderBlock(dims[1], dims[0])

    def forward(self, x):
        s0, s1, s2, s3 = x

        x = self.d3(s3, s2)
        x = self.d2(x, s1)
        x = self.d1(x, s0)
        return x


class DecoderBlock(nn.Module):
    def __init__(self, cin, cout):
        super().__init__()
        self.upsample = nn.ConvTranspose2d(cin, cout, 2, stride=2)
        self.conv = ResConv(cout*2, cout, 1)

    def forward(self, x1, x2):
        x1 = self.upsample(x1)
        x = torch.cat([x1, x2], dim=1)
        return self.conv(x)


class Bridger(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        outputs = [nn.Identity()(i) for i in x]
        return outputs


class Stem(nn.Module):
    def __init__(self, cin, cout):
        super().__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(cin, cout, 3, padding=autopad(3)),
            nn.BatchNorm2d(cout),
            nn.ReLU(),
            nn.Conv2d(cout, cout, 3, padding=autopad(3)),
        )
        self.skip = nn.Sequential(
            nn.Conv2d(cin, cout, 3, padding=autopad(3)),
        )

    def forward(self, x):
        return self.conv(x) + self.skip(x)


class ResConv(nn.Module):
    def __init__(self, cin, cout, stride=1):
        super().__init__()

        self.conv = nn.Sequential(
            nn.BatchNorm2d(cin),
            nn.ReLU(),
            nn.Conv2d(cin, cout, 3, stride, padding=autopad(3)),
            nn.BatchNorm2d(cout),
            nn.ReLU(),
            nn.Conv2d(cout, cout, 3, padding=autopad(3)),
        )
        self.skip = nn.Sequential(
            nn.Conv2d(cin, cout, 3, stride=stride, padding=autopad(3)),
            nn.BatchNorm2d(cout),
        )

    def forward(self, x):
        return self.conv(x) + self.skip(x)


class Head(nn.Sequential):
    def __init__(self, cin, cout):
        """
        Basic conv head
        :param int cin: input channel
        :param int cout: output channel
        """
        conv = nn.Conv2d(cin, cout, 1)
        super().__init__(conv)

if __name__ == '__main__':
    H = W = 224
    cout = 64
    inputs = torch.normal(0, 1, (1, 3, H, W))

    model = ResUNet(3, 100, cout)
    output = model(inputs)
    print(output.shape)

    dblock = DecoderBlock(64, 128)
    inputs2 = torch.normal(0, 1, (1, 128, H, W))
    inputs1 = torch.normal(0, 1, (1, 64, H//2, W//2))
    outputs = dblock(inputs1, inputs2)
    print(outputs.shape)

    encoder = Encoder(3, cout)
    outputs = encoder(inputs)
    for o in outputs:
        print(o.shape)

    decoder = Decoder(cout)
    outputs = decoder(outputs)
    print("Encoder + Decoder ", outputs.shape)

    stem = Stem(3, cout)
    outputs = stem(inputs)
    print(outputs.shape)
    assert list(outputs.shape) == [1, cout, H, W]

    resconv = ResConv(3, cout, 1)
    outputs = resconv(inputs)
    print(outputs.shape)
    assert list(outputs.shape) == [1, cout, H, W]

    resconv = ResConv(3, cout, stride=2)
    outputs = resconv(inputs)
    print(outputs.shape)
    assert list(outputs.shape) == [1, cout, H//2, W//2]
